@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2016-2019 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2016-2020 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -22,6 +22,7 @@
 import enum
 import itertools
 import typing
+import functools
 
 import attr
 from PyQt5.QtCore import (pyqtSignal, pyqtSlot, QUrl, QObject, QSizeF, Qt,
@@ -31,6 +32,10 @@ from PyQt5.QtWidgets import QWidget, QApplication, QDialog
 from PyQt5.QtPrintSupport import QPrintDialog, QPrinter
 from PyQt5.QtNetwork import QNetworkAccessManager
 
+if typing.TYPE_CHECKING:
+    from PyQt5.QtWebKit import QWebHistory
+    from PyQt5.QtWebEngineWidgets import QWebEngineHistory
+
 import pygments
 import pygments.lexers
 import pygments.formatters
@@ -39,12 +44,11 @@ from qutebrowser.keyinput import modeman
 from qutebrowser.config import config
 from qutebrowser.utils import (utils, objreg, usertypes, log, qtutils,
                                urlutils, message)
-from qutebrowser.misc import miscwidgets, objects
-from qutebrowser.browser import eventfilter, hints
+from qutebrowser.misc import miscwidgets, objects, sessions
+from qutebrowser.browser import eventfilter
 from qutebrowser.qt import sip
 
 if typing.TYPE_CHECKING:
-    # pylint: disable=unused-import,useless-suppression
     from qutebrowser.browser import webelem
     from qutebrowser.browser.inspector import AbstractWebInspector
 
@@ -127,11 +131,13 @@ class TabData:
     inspector = attr.ib(None)  # type: typing.Optional[AbstractWebInspector]
     open_target = attr.ib(
         usertypes.ClickTarget.normal)  # type: usertypes.ClickTarget
-    override_target = attr.ib(None)  # type: usertypes.ClickTarget
+    override_target = attr.ib(
+        None)  # type: typing.Optional[usertypes.ClickTarget]
     pinned = attr.ib(False)  # type: bool
     fullscreen = attr.ib(False)  # type: bool
     netrc_used = attr.ib(False)  # type: bool
     input_mode = attr.ib(usertypes.KeyMode.normal)  # type: usertypes.KeyMode
+    last_navigation = attr.ib(None)  # type: usertypes.NavigationRequest
 
     def should_show_icon(self) -> bool:
         return (config.val.tabs.favicons.show == 'always' or
@@ -199,7 +205,7 @@ class AbstractPrinting:
     """Attribute ``printing`` of AbstractTab for printing the page."""
 
     def __init__(self, tab: 'AbstractTab') -> None:
-        self._widget = None
+        self._widget = typing.cast(QWidget, None)
         self._tab = tab
 
     def check_pdf_support(self) -> None:
@@ -288,7 +294,7 @@ class AbstractSearch(QObject):
     def __init__(self, tab: 'AbstractTab', parent: QWidget = None):
         super().__init__(parent)
         self._tab = tab
-        self._widget = None
+        self._widget = typing.cast(QWidget, None)
         self.text = None  # type: typing.Optional[str]
         self.search_displayed = False
 
@@ -311,6 +317,7 @@ class AbstractSearch(QObject):
     def search(self, text: str, *,
                ignore_case: usertypes.IgnoreCase = usertypes.IgnoreCase.never,
                reverse: bool = False,
+               wrap: bool = True,
                result_cb: _Callback = None) -> None:
         """Find the given text on the page.
 
@@ -318,6 +325,7 @@ class AbstractSearch(QObject):
             text: The text to search for.
             ignore_case: Search case-insensitively.
             reverse: Reverse search direction.
+            wrap: Allow wrapping at the top or bottom of the page.
             result_cb: Called with a bool indicating whether a match was found.
         """
         raise NotImplementedError
@@ -350,7 +358,7 @@ class AbstractZoom(QObject):
     def __init__(self, tab: 'AbstractTab', parent: QWidget = None) -> None:
         super().__init__(parent)
         self._tab = tab
-        self._widget = None
+        self._widget = typing.cast(QWidget, None)
         # Whether zoom was changed from the default.
         self._default_zoom_changed = False
         self._init_neighborlist()
@@ -371,17 +379,18 @@ class AbstractZoom(QObject):
         It is a NeighborList with the zoom levels."""
         levels = config.val.zoom.levels
         self._neighborlist = usertypes.NeighborList(
-            levels, mode=usertypes.NeighborList.Modes.edge)
+            levels, mode=usertypes.NeighborList.Modes.edge
+        )  # type: usertypes.NeighborList[float]
         self._neighborlist.fuzzyval = config.val.zoom.default
 
-    def apply_offset(self, offset: int) -> None:
+    def apply_offset(self, offset: int) -> float:
         """Increase/Decrease the zoom level by the given offset.
 
         Args:
             offset: The offset in the zoom level list.
 
         Return:
-            The new zoom percentage.
+            The new zoom level.
         """
         level = self._neighborlist.getitem(offset)
         self.set_factor(float(level) / 100, fuzzyval=False)
@@ -434,7 +443,7 @@ class AbstractCaret(QObject):
                  parent: QWidget = None) -> None:
         super().__init__(parent)
         self._tab = tab
-        self._widget = None
+        self._widget = typing.cast(QWidget, None)
         self.selection_enabled = False
         self._mode_manager = mode_manager
         mode_manager.entered.connect(self._on_mode_entered)
@@ -527,7 +536,7 @@ class AbstractScroller(QObject):
     def __init__(self, tab: 'AbstractTab', parent: QWidget = None):
         super().__init__(parent)
         self._tab = tab
-        self._widget = None  # type: typing.Optional[QWidget]
+        self._widget = typing.cast(QWidget, None)
         if 'log-scroll-pos' in objects.debug_flags:
             self.perc_changed.connect(self._log_scroll_pos_change)
 
@@ -597,7 +606,8 @@ class AbstractHistoryPrivate:
 
     def __init__(self, tab: 'AbstractTab'):
         self._tab = tab
-        self._history = None
+        self._history = typing.cast(
+            typing.Union['QWebHistory', 'QWebEngineHistory'], None)
 
     def serialize(self) -> bytes:
         """Serialize into an opaque format understood by self.deserialize."""
@@ -618,7 +628,8 @@ class AbstractHistory:
 
     def __init__(self, tab: 'AbstractTab') -> None:
         self._tab = tab
-        self._history = None
+        self._history = typing.cast(
+            typing.Union['QWebHistory', 'QWebEngineHistory'], None)
         self.private_api = AbstractHistoryPrivate(tab)
 
     def __len__(self) -> int:
@@ -679,7 +690,7 @@ class AbstractElements:
     _ErrorCallback = typing.Callable[[Exception], None]
 
     def __init__(self, tab: 'AbstractTab') -> None:
-        self._widget = None
+        self._widget = typing.cast(QWidget, None)
         self._tab = tab
 
     def find_css(self, selector: str,
@@ -740,7 +751,7 @@ class AbstractAudio(QObject):
 
     def __init__(self, tab: 'AbstractTab', parent: QWidget = None) -> None:
         super().__init__(parent)
-        self._widget = None  # type: typing.Optional[QWidget]
+        self._widget = typing.cast(QWidget, None)
         self._tab = tab
 
     def set_muted(self, muted: bool, override: bool = False) -> None:
@@ -771,7 +782,7 @@ class AbstractTabPrivate:
 
     def __init__(self, mode_manager: modeman.ModeManager,
                  tab: 'AbstractTab') -> None:
-        self._widget = None  # type: typing.Optional[QWidget]
+        self._widget = typing.cast(QWidget, None)
         self._tab = tab
         self._mode_manager = mode_manager
 
@@ -812,14 +823,6 @@ class AbstractTabPrivate:
         """
         raise NotImplementedError
 
-    def user_agent(self) -> typing.Optional[str]:
-        """Get the user agent for this tab.
-
-        This is only implemented for QtWebKit.
-        For QtWebEngine, always returns None.
-        """
-        raise NotImplementedError
-
     def shutdown(self) -> None:
         raise NotImplementedError
 
@@ -881,19 +884,18 @@ class AbstractTab(QWidget):
 
         self.data = TabData()
         self._layout = miscwidgets.WrapperLayout(self)
-        self._widget = None  # type: typing.Optional[QWidget]
+        self._widget = typing.cast(QWidget, None)
         self._progress = 0
         self._has_ssl_errors = False
         self._load_status = usertypes.LoadStatus.none
         self._tab_event_filter = eventfilter.TabEventFilter(
             self, parent=self)
-        self.backend = None
+        self.backend = None  # type: typing.Optional[usertypes.Backend]
 
-        # FIXME:qtwebengine  Should this be public api via self.hints?
-        #                    Also, should we get it out of objreg?
-        hintmanager = hints.HintManager(win_id, self.tab_id, parent=self)
-        objreg.register('hintmanager', hintmanager, scope='tab',
-                        window=self.win_id, tab=self.tab_id)
+        # If true, this tab has been requested to be removed (or is removed).
+        self.pending_removal = False
+        self.shutting_down.connect(functools.partial(
+            setattr, self, 'pending_removal', True))
 
         self.before_load_started.connect(self._on_before_load_started)
 
@@ -988,6 +990,9 @@ class AbstractTab(QWidget):
                                       navigation.navigation_type,
                                       navigation.is_main_frame))
 
+        if navigation.is_main_frame:
+            self.data.last_navigation = navigation
+
         if not navigation.url.isValid():
             # Also a WORKAROUND for missing IDNA 2008 support in QUrl, see
             # https://bugreports.qt.io/browse/QTBUG-60364
@@ -1011,13 +1016,9 @@ class AbstractTab(QWidget):
             # https://github.com/qutebrowser/qutebrowser/issues/3498
             return
 
-        try:
-            sess_manager = objreg.get('session-manager')
-        except KeyError:
-            # https://github.com/qutebrowser/qutebrowser/issues/4311
-            return
+        if sessions.session_manager is not None:
+            sessions.session_manager.save_autosave()
 
-        sess_manager.save_autosave()
         self.load_finished.emit(ok)
 
         if not self.title():
